#!/usr/bin/tclsh
source cgi.tcl
source json.tcl
source xml.tcl
package require msgcat
namespace import msgcat::mc

# States:
# Idle -> Download -> Install +-> Success
#                             +-> Error

# STATUS_FILE format:
# <State>|<Percent>
# (if State is 'Install' use GDSUPDATE_PROGRESSFILE)
set env(LD_LIBRARY_PATH) /opt/gira/lib
set TMPDIR  "/opt/userdata/tmp/fwupdate"
set TMP_FIRMWARE_FILE "${TMPDIR}/new_device_package.zip"
set HOST 127.0.0.1
#set HOST 192.168.1.193
set STATUS_FILE "/var/tmp/fwupdate_status"
set GDSUPDATE "/opt/gira/bin/gdsupdate"
set GDSURL "https://${HOST}:4433/xmlrpc"
set GDSUPDATE_PROGRESSFILE "/var/tmp/gdsupdate_progress"
set GDSUPDATE_LOGFILE "/var/tmp/gdsupdate_log"
set POLL_INTERVAL 5
set DESCRIPTION_URL "http://${HOST}:8080/DeviceDescription.xml"

set INFO_URL_FORMAT "http://www.ise.de/system/fwu/%s.xml"

proc cgi_cgi {args} {return $args}

set _cgi(tmpdir) $TMPDIR

file mkdir $TMPDIR

proc get_device_description {} {
    global DESCRIPTION_URL
    set xml [exec wget $DESCRIPTION_URL -q -O -]
    return [xml2list $xml]
}

proc get_device_info { model_var firmware_version_var name_var manufacturer_var } {
    upvar $model_var model 
    upvar $firmware_version_var firmware_version 
    upvar $name_var name
    upvar $manufacturer_var manufacturer
    set document [get_device_description]
    set description [xml_get_node $document {Description}]
    set firmware_version [xml_get_attribute $description {FirmwareVersion Value}]
    set model [xml_get_attribute $description {Model Value}]
    set name [xml_get_attribute $description {Model Name}]
    set manufacturer [xml_get_attribute $description {Manufacturer Value}]
}

proc version_compare { v1 v2 } {
    set v1_list [split $v1 .]
    set v2_list [split $v2 .]
    set i 0
    for { set i 0 } { ($i < [llength $v1_list]) && ($i < [llength $v2_list]) } { incr i } {
        set v1_component [lindex $v1_list $i]
        set v2_component [lindex $v2_list $i]
        if { $v1_component < $v2_component } { return -1 }
        if { $v1_component > $v2_component } { return 1 }
    }
    if { $i < [llength $v1_list] } { return 1 }
    if { $i < [llength $v2_list] } { return -1 }
    return 0
}

proc make_url_absolute { ref_url url } {
    if { [regexp {^[a-z]+://} $url] } {
        return $url
    }
    if { [regexp {^/} $url] } {
        regexp {^[a-z]+://[^/]+} $ref_url base_url
        return "${base_url}${url}"
    }
    if { [regexp {^./} $url] } {
        set url [string range $url 2 end]
    }
    if { ![regexp {(^[a-z]+://.*/)[^/]*} $ref_url dummy base_url] } {
        return "${ref_url}/${url}"
    }
    return "${base_url}${url}"
}

proc get_latest_firmware { deviceid lang array_var } {
    global INFO_URL_FORMAT
    upvar $array_var fwinfo
    set info_url [format $INFO_URL_FORMAT $deviceid]
    set xml [exec wget --no-check-certificate $info_url -q -O -]
    set document [xml2list $xml]
    
    set firmware_nodes [lindex $document 2]
    set latest_version ""
    foreach firmware_node $firmware_nodes {
        if { [llength $firmware_node] != 3 } {
            error "Expected list of 3 elements but got [llength $n]"
        }
        if { "[lindex $firmware_node 0]" != "Firmware" } {
            continue
        }
        set firmware_version [xml_get_attribute $firmware_node "Version"]
        if { [version_compare $firmware_version $latest_version] > 0 } {
            set latest_version $firmware_version
            array set fwinfo [lindex $firmware_node 1]
            set fwinfo(DevicePackageUrl) [make_url_absolute $info_url $fwinfo(DevicePackageUrl)]
            foreach description_node [lindex $firmware_node 2] {
               
                array set descr_attrs [lindex $description_node 1]
                if { ![info exist descr_attrs(Language)] } {
                    if { ![info exist fwinfo(Description)] } {
                        set fwinfo(Description) [xml_get_text $description_node]
                    }
                } elseif { "[string tolower $lang]" == "[string tolower $descr_attrs(Language)]" } {
                    set fwinfo(Description) [xml_get_text $description_node]
                }
            }
            
        }
    }
    return [info exist fwinfo(DevicePackageUrl)]
}

proc read_status { percent_var msg_var } {
    global STATUS_FILE
    upvar $percent_var percent
    upvar $msg_var msg
    if { [catch {
        set fd [open $STATUS_FILE r]
        }]} {
        set percent 0
        set msg ""
        return Idle
    }
    set line [gets $fd]
    close $fd
    set fields [split $line "|"]
    set status [lindex $fields 0]
    set percent [lindex $fields 1]
    set msg [lindex $fields 2]
    return $status
    
}

proc write_status {status percent {msg ""} } {
    global STATUS_FILE
    set fd [open ${STATUS_FILE}.new w]
    puts $fd $status|$percent|$msg
    close $fd
    file rename -force -- ${STATUS_FILE}.new ${STATUS_FILE}
}

proc show_form {} {
    global env
    
    cgi_input

    set vars [cgi_cookie_list]
    if { [lsearch $vars "isc_language"] >= 0 } {
        cgi_import_cookie isc_language
    }
    
    set local_update 1
    catch { 
        cgi_import local_update 
        set local_update 1
    }
    if {[info exists isc_language]} {
        set lang $isc_language
    } else {
        set lang [cgi_language]
        catch { 
            cgi_import lang
        }
    }
    msgcat::mclocale $lang
    msgcat::mcload [file join [file dirname [info script]] msgs]


    http_head

    get_device_info model cur_firmware_version model_name manufacturer
    set deviceid $manufacturer
    append deviceid $model

    html {
        cgi_head {
            title $model_name
            relationship stylesheet "/style.css" "type=text/css"

            
            script "type=text/javascript" "src=/iscui/jquery/jquery-1.12.0.min.js" {}
            script "type=text/javascript" "src=/iscui/jquery/jquery.cookie.js" {}
            script "type=text/javascript" "src=/iscui/jquery/jquery.form.min.js" {}
            script "type=text/javascript" "src=js/script.js" {}
            script "type=text/javascript" "src=/iscui/i18n.js" {}
            script "type=text/javascript" "src=/iscui/lang.js" {}
            script "type=text/javascript" "src=/kim.js" {}
			script "type=text/javascript" "src=/index.js" {}
			script "type=text/javascript" "src=/App.js" {}
            script "type=text/javascript" {
                cgi_puts "var lang = \"$lang\";"
                cgi_puts "var msgStartingUpdate = \"[mc {Starting update}]\";"
                cgi_puts "var msgDownloadingDevicePackage = \"[mc {Downloading firmware}]\";"
                cgi_puts "var msgUploadingDevicePackage = \"[mc {Uploading firmware}]\";"
                cgi_puts "var msgUploadingDevicePackageComplete = \"[mc {Uploading firmware - Complete}]\";"
                cgi_puts "var msgUpdatingFiles = \"[mc {Updating files}]\";"
                cgi_puts "var msgRestart = \"[mc {Device is being restarted. This page will be reloaded when the device is ready again.}]\";"
                cgi_puts "var msgProcessingDevicePackage = \"[mc {Firmware file is being processed.}]\";"
            }
        }
    
        body "onload=\"InitializeLanguage(); FillCommonElements(); createTopNavMenu();\"" {
            division "id=header" {
                cgi_puts [url $model_name / {id="friendlyDeviceNameHeading" style="font-size:30px; font-family:'Trebuchet MS', Arial, Verdana, sans-serif; position:relative; top:25px; left:25px; color:#fff; text-decoration:none;"}]
                cgi_puts [img /logo.png {style="position:absolute; top:35px; right:13px; width:102px; height:58px;"}]
            }
            division "id=topnav" {
                bullet_list {
                    li {}
                }
                division "id=language_switch" {}
            }
            division "id=middle" {
				division "id=modal" {
				}
			
                set state [read_status percent message]
        
                if { "$state" == "Idle" } {
                    h1 [mc "Online update"]
                    h2 [mc "Current firmware version: %s" $cur_firmware_version]
        
                    if { [catch {
                        if { [get_latest_firmware $deviceid $lang fwinfo] } {
                            if { [version_compare $cur_firmware_version $fwinfo(Version)] >= 0 } {
                                h2 [mc "No newer firmware available"]
                            } else {
                                h2 [mc "A firmware update to version %s is available" $fwinfo(Version)]
                                if { [info exist fwinfo(Description)] } {
                                    cgi_p $fwinfo(Description)
                                }
                                set cur_main_version [lindex [split $cur_firmware_version .] 0]
                                set cur_sub_version [lindex [split $cur_firmware_version .] 1]
                                set new_main_version [lindex [split $fwinfo(Version) .] 0]
                                set yes_no_message ""
                                if { [info exist fwinfo(BreakingChangeToVersion)] && ( $fwinfo(BreakingChangeToVersion) > 0 ) && ( $fwinfo(BreakingChangeToVersion) > "$cur_main_version.$cur_sub_version" ) } {
                                    set warn_message [mc {This firmware update provides new functionality which is incompatible to the current configuration. If you perform this update, the device must be re-configured with ETS to work again properly!}]
                                    cgi_br
                                    cgi_p "class=warning" $warn_message
                                    set yes_no_message "$warn_message\\n\\n[mc {Do you really want to update the firmware?}]"
                                } elseif { $cur_main_version < $new_main_version } {
                                    set info_message [mc "This firmware update provides new functionality but is backward compatible to the current configuration. However, if you want to update the configuration via ETS, you have to upgrade the ETS Application Program to the latest version in your ETS project."]
                                    cgi_br
                                    cgi_p "class=info" $info_message
                                    set yes_no_message "$info_message\\n\\n[mc {Do you really want to update the firmware?}]"
                                }
                                cgi_br
                                button [mc "Perform update"] "id=button_update" "onclick=startServerUpdate('$yes_no_message')"
                            }
                        } else {
                            h2 [mc "No firmware available"]
                        }
                    }]} {
                        h2 [mc "The firmware update server is currently not reachable"]
                    }
                    
                    if { $local_update } {
                        cgi_br
                        cgi_br
                        h1 [mc "Local update without internet access"]
                        h2 [mc "Upload firmware file"]
                        form  $env(SCRIPT_NAME) enctype=multipart/form-data method=post {
                            division "class=upload" {
                                cgi_export lang=[msgcat::mclocale]
                                file_button "firmware_file"
                            }
                            cgi_br
                            submit_button "=[mc {Perform update}]"
                        }
                    }
                }

                cgi_br
                cgi_br
                h1 {id="heading_progress"} [mc "Firmware update progress"]

                division "class=progress" "id=install_progress" {
                    division "class=bar" {}
                    division "class=percent" {}
                }
                button [mc "Reset error"] "id=button_reset" "onclick=resetAfterError()"
            }
            
            division "id=footer" {
            }
            
            if { "$state" != "Idle" } {
                script "type=text/javascript" {
                    cgi_puts {
                        $('#heading_progress').show();
                        pollStatus( "", -1 );
                    }
                }
            }
            
        }
    }
}

proc puts_progress { state progress message } {
    set message [escape_json_string $message]
    cgi_puts "{ \"progress\": $progress, \"state\": \"$state\", \"message\": \"$message\", \"lang\": \"[::msgcat::mclocale]\" }"
}


proc handle_upload {} {
    global env TMP_FIRMWARE_FILE
    file mkdir [file dirname $TMP_FIRMWARE_FILE]
    cgi_input
    
    set lang [cgi_language]
    catch { import lang }
    msgcat::mclocale $lang
    msgcat::mcload [file join [file dirname [info script]] msgs]
    
    import_file -client firmware_file
    file rename -force -- [lindex $firmware_file 0] $TMP_FIRMWARE_FILE
    content_type "application/json"
    
    puts_progress Upload 100 [mc "Upload complete"]
    
}

proc handle_json {} {
    global env
    set length $env(CONTENT_LENGTH)
    if {0!=[string compare $length "-1"]} {
        if { $length } {
            set lang [cgi_language]
            set input [read stdin $env(CONTENT_LENGTH)]
            array set request [parse_json $input]
            if [info exist request(lang)] {
                set lang $request(lang)
            }
            msgcat::mclocale $lang
            msgcat::mcload [file join [file dirname [info script]] msgs]
            switch $request(cmd) {
                download {
                    download_start
                }
                install {
                    update_start
                }
                reset {
                    reset_state
                }
                poll {
                    if { !([info exist request(last_state)] && [info exist request(last_progress)]) } {
                        update_poll "" -1
                    } else {
                        update_poll $request(last_state) $request(last_progress)
                    }
                }
                default {
                    content_type "application/json"
                    puts_progress Error 0 [mc "Invalid command %s" $request(cmd)]
                }
            }
        }
    }
}

proc start_gdsupdate {} {
    global TMP_FIRMWARE_FILE GDSUPDATE GDSURL GDSUPDATE_PROGRESSFILE GDSUPDATE_LOGFILE
    
    file delete $GDSUPDATE_PROGRESSFILE
    
    if { [catch {
        set pid [exec $GDSUPDATE -p $TMP_FIRMWARE_FILE -u $GDSURL -o $GDSUPDATE_PROGRESSFILE >& $GDSUPDATE_LOGFILE &]
        write_status Install 0 $pid
    } err] } {
        write_status Error 0 $err
    }
}

proc update_start {} {
    start_gdsupdate
    update_poll 0
}

proc reset_state {} {
    global TMP_FIRMWARE_FILE GDSUPDATE_PROGRESSFILE STATUS_FILE
    file delete $STATUS_FILE
    file delete $GDSUPDATE_PROGRESSFILE
    file delete $TMP_FIRMWARE_FILE
    content_type "application/json"
    cgi_puts "{ \"status\": true }"
}

proc download_start {} {
    global argv0
    
    get_device_info model cur_firmware_version model_name manufacturer
    set deviceid $manufacturer
    append deviceid $model
    if { ![get_latest_firmware $deviceid "" fwinfo] } {
        content_type "application/json"
        puts_progress Error 0 [mc "No firmware available"]
        return
    }
    write_status Download 0
    exec [info nameofexecutable] $argv0 $fwinfo(DevicePackageUrl) >& /dev/null &
    update_poll
}

proc update_poll { {last_state ""} {last_percent -1} } {
    global GDSUPDATE_PROGRESSFILE POLL_INTERVAL
    
    content_type "application/json"
    
    set seconds 0
    while { $seconds < $POLL_INTERVAL } {
        set state [read_status percent message]
        if { "$state" == "Install" } {
            if { "$state" != "$last_state" } {
                set percent 0
            } else {
                set percent $last_percent
            }

            set gdsupdate_pid $message
            if { ![info exist fd] } {
                catch {
                    set fd [open $GDSUPDATE_PROGRESSFILE r]
                }
            }
            if { [info exist fd] } {
                while { [gets $fd line] > 0 } {
                    set progress [split $line "|"]
                    if { "[lindex $progress 0]" == "Progress" } {
                        set percent [lindex $progress 1]
                    } else {
                        set state [lindex $progress 0]
                    }
                    set message [lindex $progress 2]
                }
            }
            if { ("$state" != "Error") && ("$state" != "Success") } {
                if { ![file exists /proc/${gdsupdate_pid}/cmdline] } {
                    set state Error
                    set progress 0
                    set message  [mc "Update failed"]
                    write_status $state $progress $message
                }
            }
        }
        if { ("$state" != "$last_state") || ("$percent" != "$last_percent") } {
            break
        }
        after 1000
        incr seconds
    }
    
    switch $state {
        Download {
            set message [mc "Downloading firmware"]
        }
        Success {
            set message [mc "Rebooting"]
        }
        Error {
            if { [info exist message] } {
                set message [mc "Error: %s" $message]
            } else {
                set message [mc "Error"]
            }
        }
        default {
            if { ![info exist message] } {
                set message ""
            }
        }
    }
    
    if { $percent < 0 } { set percent 0 }
    puts_progress $state $percent $message
    
    if { [info exist fd] } {
        close $fd
    }
}

proc server_update { url } {
    global TMP_FIRMWARE_FILE
    if { [catch {
        file mkdir [file dirname $TMP_FIRMWARE_FILE]
        set file_info [exec wget --no-check-certificate --spider $url 2>@1]
        if { ![regexp {^Length: ([0-9]+) } $file_info dummy length] } {
            set length 20000000
        }
        set BUFSIZE 65536
        set wget_fd [open "|wget  --no-check-certificate -q -O - $url" r]
        set dp_fd [open $TMP_FIRMWARE_FILE w]
        fconfigure $wget_fd -translation binary -blocking 1
        fconfigure $dp_fd -translation binary -blocking 1
        set eof 0
        set percent 0
        set count 0
        write_status Download $percent
        while { 1 } {
            set buffer [read $wget_fd $BUFSIZE]
            if { ![string length $buffer] } {
                break
            }
            puts -nonewline $dp_fd $buffer
            incr count [string length $buffer]
            set percent [expr int($count * 100 / $length) ]
            if { $percent > 99 } { set percent 99 }
            write_status Download $percent
        }
        catch { close $dp_fd }
        write_status Download 100
    
        start_gdsupdate
    } msg] } {
        write_status Error 0 "$msg"
    }
}

if { $argc > 0 } {
    server_update [lindex $argv 0]
    exit 0
}

cgi_eval {
    if [file exists /opt/userdata/.cgi-debug] {cgi_debug -on}

    if { $env(REQUEST_METHOD) == "GET" } {
        show_form
    } elseif { $env(REQUEST_METHOD) == "POST" } {
        if {[info exists env(CONTENT_TYPE)] && [regexp ^multipart/form-data $env(CONTENT_TYPE)]} {
            handle_upload
        } else {
            handle_json
        }
    }
}

